using System;
using System.Collections.Generic;
using System.Windows.Forms;
using System.Xml;
using System.IO;

using VVVV.PluginInterfaces.V1;

namespace Hoster
{
	public partial class MainForm
	{
		public List<PlugInfo> FPlugins;
		private XmlDocument FDocument;
		
		[STAThread]
		public static void Main(string[] args)
		{
			Application.EnableVisualStyles();
			Application.SetCompatibleTextRenderingDefault(false);
			Application.Run(new MainForm());
		}
		
		public MainForm()
		{
			// The InitializeComponent() call is required for Windows Forms designer support.
			InitializeComponent();
			
			FPlugins = new List<PlugInfo>();
			ScanForPlugins();

			if (! System.IO.File.Exists("./savegame.xml"))
			{
				XmlTextWriter xml = new XmlTextWriter("savegame.xml", null);
				xml.WriteStartElement("HOST");
				xml.WriteEndElement();
				xml.Close();
			}
			
			FDocument = new XmlDocument();
			LoadFile();
			
			
			if (PluginPages.TabPages.Count == 0)
				AddSelektorPage();
		}
		
		public void SetPlugin(TabPage page, int PlugIndex)
		{
			RemovePage(page);
			
			AddPluginPage(FPlugins[PlugIndex].Path, FPlugins[PlugIndex].ClassName);
		}
		
		public void RemovePage(TabPage page)
		{
			PluginPages.TabPages.Remove(page);
			page.Dispose();
		}
		
		private PluginPage AddPluginPage(string Path, string ClassName)
		{
			TabPage tab = new TabPage();
			tab.Text = ClassName;
			PluginPages.TabPages.Add(tab);
			tab.Height = PluginPages.Height; //initializing
			PluginPages.SelectedIndex = PluginPages.TabPages.Count-1;
			
			PluginPage pp = new PluginPage();
			pp.Parent = tab;
			pp.Dock = DockStyle.Fill;
			
			pp.LoadPlugin(Path, ClassName);
			
			return pp;
		}
		
		private void AddSelektorPage()
		{
			TabPage tab = new TabPage();
			SelektorPage sp = new SelektorPage(this);
			sp.Dock = DockStyle.Fill;
			sp.Parent = tab;
			tab.Text = "PluginSelektor";

			PluginPages.TabPages.Add(tab);
			PluginPages.SelectedIndex = PluginPages.TabPages.Count-1;
		}
		
		void ScanForPlugins()
		{
			string[] dlls;
			Type objInterface;
			string path = Application.StartupPath;
			dlls = System.IO.Directory.GetFileSystemEntries(path, "*.dll");
			
			System.Reflection.Assembly plugindll;
			for (int i=0; i<dlls.Length; i++)
			{
				try
				{
					plugindll = System.Reflection.Assembly.LoadFrom(dlls[i]);
					//Loop through each type in the DLL
					
					foreach (System.Type objType in plugindll.GetTypes())
					{
						//Only look at public, non abstract types
						if (objType.IsPublic && !objType.IsAbstract)
						{
							//See if this type implements our interface
							objInterface = objType.GetInterface("VVVV.PluginInterfaces.V1.IPlugin");
							if (objInterface != null)
							{
								PlugInfo p = new PlugInfo();
								p.ClassName = objType.FullName;
								p.Path = plugindll.Location;
								FPlugins.Add(p);
							}
						}
					}
				}
				catch
				{
					//some .dlls except on being asked for types
					//Console.Beep(400, 10);
				}
			}
		}
		
		void PluginPagesMouseClick(object sender, System.Windows.Forms.MouseEventArgs e)
		{
			if (e.Button == MouseButtons.Middle)
			{
				for (int i=0; i<PluginPages.TabPages.Count; i++)
				{
					if (PluginPages.GetTabRect(i).Contains(e.Location))
					{
						RemovePage(PluginPages.TabPages[i]);
						break;
					}
				}
			}

			if (PluginPages.TabPages.Count == 0)
				AddSelektorPage();
		}
		
		void Panel2MouseDoubleClick(object sender, System.Windows.Forms.MouseEventArgs e)
		{
			if (e.Button == MouseButtons.Left)
			{
				AddSelektorPage();
			}
		}

		void SaveButtonClick(object sender, System.EventArgs e)
		{
			Save();
		}
		
		void LoadButtonClick(object sender, System.EventArgs e)
		{
			//Load();
		}
		
		void Save()
		{
			string path = Path.Combine(Application.StartupPath, "savegame.xml");
			FDocument.Load(path);
			XmlNode host = FDocument.SelectSingleNode(@"//HOST");
			if (host != null)
			{
				host.RemoveAll();
				
				for (int i=0; i<PluginPages.TabPages.Count; i++)
				{
					if (PluginPages.TabPages[i].Controls[0] is PluginPage)
						(PluginPages.TabPages[i].Controls[0] as PluginPage).SaveToXML(FDocument);
				}
				
				//save host attributes
				XmlAttribute st = FDocument.CreateAttribute("selectedtab");
				st.Value = PluginPages.SelectedIndex.ToString();
				host.Attributes.Append(st);
				st = FDocument.CreateAttribute("windowleft");
				st.Value = this.Left.ToString();
				host.Attributes.Append(st);
				st = FDocument.CreateAttribute("windowtop");
				st.Value = this.Top.ToString();
				host.Attributes.Append(st);
				st = FDocument.CreateAttribute("windowwidth");
				st.Value = this.Width.ToString();
				host.Attributes.Append(st);
				st = FDocument.CreateAttribute("windowheight");
				st.Value = this.Height.ToString();
				host.Attributes.Append(st);
			}
			
			FDocument.Save(path);
		}
		
		void LoadFile()
		{
			string path = Path.Combine(Application.StartupPath, "savegame.xml");
			FDocument.Load(path);
			
			XmlNode host;
			XmlAttribute attr;
			string[] nodename;
			char[] separator = {'|'};
			host = FDocument.SelectSingleNode(@"//HOST");
			
			if (host != null)
			{
				//restore window
				attr = host.Attributes.GetNamedItem("windowleft") as XmlAttribute;
				if (attr != null)
					this.Left = Convert.ToInt32(attr.Value);
				attr = host.Attributes.GetNamedItem("windowtop") as XmlAttribute;
				if (attr != null)
					this.Top = Convert.ToInt32(attr.Value);
				attr = host.Attributes.GetNamedItem("windowwidth") as XmlAttribute;
				if (attr != null)
					this.Width = Convert.ToInt32(attr.Value);
				attr = host.Attributes.GetNamedItem("windowheight") as XmlAttribute;
				if (attr != null)
					this.Height = Convert.ToInt32(attr.Value);
				
				foreach(XmlNode plugin in host.ChildNodes)
				{
					attr = plugin.LastChild.Attributes.GetNamedItem("nodename") as XmlAttribute;
					nodename = attr.Value.Split(separator);
					
					if (System.IO.File.Exists(nodename[0]))
					{
						PluginPage pp = AddPluginPage(nodename[0], nodename[1]);
						System.Diagnostics.Debug.WriteLine("Loading From XML");
						pp.LoadFromXML(plugin);
					}
				}
				
				//select saved tabindex
				attr = host.Attributes.GetNamedItem("selectedtab") as XmlAttribute;
				if (attr != null)
					PluginPages.SelectedIndex = Convert.ToInt32(attr.Value);
			}
		}
	}
	
	public class PlugInfo
	{
		private string FPath;
		private string FClassName;
		
		public string Path
		{
			get {return FPath;}
			set {FPath = value;}
		}
		public string ClassName
		{
			get {return FClassName;}
			set {FClassName = value;}
		}
	}
}
