#region licence/info

//////project name
//vvvv plugin template

//////description
//basic vvvv node plugin template.
//Copy this an rename it, to write your own plugin node.

//////licence
//GNU Lesser General Public License (LGPL)
//english: http://www.gnu.org/licenses/lgpl.html
//german: http://www.gnu.de/lgpl-ger.html

//////language/ide
//C# sharpdevelop

//////dependencies
//VVVV.PluginInterfaces.V1;
//VVVV.Utils.VColor;
//VVVV.Utils.VMath;

//////initial author
//vvvv group

#endregion licence/info

//use what you need
using System;
using System.Runtime.InteropServices;
using System.Drawing;
using System.Collections.Generic;
using System.Collections;

using VVVV.PluginInterfaces.V1;
using VVVV.Utils.VColor;
using VVVV.Utils.VMath;

using SlimDX.Direct3D9;
using SlimDX;

//the vvvv node namespace
namespace VVVV.Nodes
{
	//class definition
	public class PluginMeshTemplate: IPlugin, IDisposable
	{
		#region field declaration
		
		//the host (mandatory)
		private IPluginHost FHost;
		//Track whether Dispose has been called.
		private bool FDisposed = false;


        //a mesh output pin
        private IValueIn FInputVertices, FInputNormals, FInputIndices, FInputTexCoords;
        private IValueOut FOutputVertices, FOutputNormals, FOutputOriginalTexCoords, FOutputIndices, FOutputTexCoords;
		
		//a list that holds a mesh for every device
		private Dictionary<int, Mesh> FDeviceMeshes = new Dictionary<int, Mesh>();
		
		#endregion field declaration
		
		#region constructor/destructor
		
		public PluginMeshTemplate()
		{
			//the nodes constructor
			//nothing to declare for this node
		}
		
		// Implementing IDisposable's Dispose method.
		// Do not make this method virtual.
		// A derived class should not be able to override this method.
		public void Dispose()
		{
			Dispose(true);
			// Take yourself off the Finalization queue
			// to prevent finalization code for this object
			// from executing a second time.
			GC.SuppressFinalize(this);
		}
		
		// Dispose(bool disposing) executes in two distinct scenarios.
		// If disposing equals true, the method has been called directly
		// or indirectly by a user's code. Managed and unmanaged resources
		// can be disposed.
		// If disposing equals false, the method has been called by the
		// runtime from inside the finalizer and you should not reference
		// other objects. Only unmanaged resources can be disposed.
		protected virtual void Dispose(bool disposing)
		{
			// Check to see if Dispose has already been called.
			if(!FDisposed)
			{
				if(disposing)
				{
					// Dispose managed resources.
					
				}
				// Release unmanaged resources. If disposing is false,
				// only the following code is executed.

				FHost.Log(TLogType.Debug, "PluginMeshTemplate is being deleted");
				
				// Note that this is not thread safe.
				// Another thread could start disposing the object
				// after the managed resources are disposed,
				// but before the disposed flag is set to true.
				// If thread safety is necessary, it must be
				// implemented by the client.
			}
			FDisposed = true;
		}

		// Use C# destructor syntax for finalization code.
		// This destructor will run only if the Dispose method
		// does not get called.
		// It gives your base class the opportunity to finalize.
		// Do not provide destructors in types derived from this class.
		~PluginMeshTemplate()
		{
			// Do not re-create Dispose clean-up code here.
			// Calling Dispose(false) is optimal in terms of
			// readability and maintainability.
			Dispose(false);
		}
		#endregion constructor/destructor
		
		#region node name and info
		
		//provide node infos
		private static IPluginInfo FPluginInfo;
		public static IPluginInfo PluginInfo
		{
			get
			{
				if (FPluginInfo == null)
				{
					//fill out nodes info
					//see: http://www.vvvv.org/tiki-index.php?page=Conventions.NodeAndPinNaming
					FPluginInfo = new PluginInfo();
					
					//the nodes main name: use CamelCaps and no spaces
					FPluginInfo.Name = "FurFinsMesh";
					//the nodes category: try to use an existing one
					FPluginInfo.Category = "Plugins";
					//the nodes version: optional. leave blank if not
					//needed to distinguish two nodes of the same name and category
					FPluginInfo.Version = "";
					
					//the nodes author: your sign
					FPluginInfo.Author = "Rene Fuerst";
					//describe the nodes function
					FPluginInfo.Help = "Creates a mesh for rendering fur fins.";
					//specify a comma separated list of tags that describe the node
					FPluginInfo.Tags = "";
					
					//give credits to thirdparty code used
					FPluginInfo.Credits = "";
					//any known problems?
					FPluginInfo.Bugs = "";
					//any known usage of the node that may cause troubles?
					FPluginInfo.Warnings = "";
					
					//leave below as is
					System.Diagnostics.StackTrace st = new System.Diagnostics.StackTrace(true);
					System.Diagnostics.StackFrame sf = st.GetFrame(0);
					System.Reflection.MethodBase method = sf.GetMethod();
					FPluginInfo.Namespace = method.DeclaringType.Namespace;
					FPluginInfo.Class = method.DeclaringType.Name;
					//leave above as is
				}
				return FPluginInfo;
			}
		}

		public bool AutoEvaluate
		{
			//return true if this node needs to calculate every frame even if nobody asks for its output
			get {return false;}
		}
		
		#endregion node name and info
		
		#region pin creation
		
		//this method is called by vvvv when the node is created
		public void SetPluginHost(IPluginHost Host)
		{
			//assign host
			FHost = Host;
			
			//create outputs
            FHost.CreateValueInput("Model vertices", 3, null, TSliceMode.Dynamic, TPinVisibility.True, out FInputVertices);
            FInputVertices.SetSubType3D(Double.MinValue, Double.MaxValue, 0.1, 0.0, 0.0, 0.0, false, false, false);

            FHost.CreateValueInput("Model normals", 3, null, TSliceMode.Dynamic, TPinVisibility.True, out FInputNormals);
            FInputNormals.SetSubType3D(Double.MinValue, Double.MaxValue, 0.1, 0.0, 0.0, 0.0, false, false, false);

            FHost.CreateValueInput("Model TexCoords", 2, null, TSliceMode.Dynamic, TPinVisibility.True, out FInputTexCoords);
            FInputTexCoords.SetSubType2D(Double.MinValue, Double.MaxValue, 0.1, 0.0, 0.0, false, false, false);
            
            FHost.CreateValueInput("Model indices", 1, null, TSliceMode.Dynamic, TPinVisibility.True, out FInputIndices);

            FHost.CreateValueOutput("Transformed vertices", 3, null, TSliceMode.Dynamic, TPinVisibility.True, out FOutputVertices);
            FOutputVertices.SetSubType3D(Double.MinValue, Double.MaxValue, 0.1, 0.0, 0.0, 0.0, false, false, false);

            FHost.CreateValueOutput("Normal information", 3, null, TSliceMode.Dynamic, TPinVisibility.True, out FOutputNormals);
            FOutputNormals.SetSubType3D(Double.MinValue, Double.MaxValue, 0.1, 0.0, 0.0, 0.0, false, false, false);
            
            FHost.CreateValueOutput("Texture coords", 2, null, TSliceMode.Dynamic, TPinVisibility.True, out FOutputTexCoords);
            FOutputTexCoords.SetSubType2D(Double.MinValue, Double.MaxValue, 0.1, 0.0, 0.0, false, false, false);

            FHost.CreateValueOutput("Original Texture coords", 2, null, TSliceMode.Dynamic, TPinVisibility.True, out FOutputOriginalTexCoords);
            FOutputOriginalTexCoords.SetSubType2D(Double.MinValue, Double.MaxValue, 0.1, 0.0, 0.0, false, false, false);

            FHost.CreateValueOutput("Transformed indices", 1, null, TSliceMode.Dynamic, TPinVisibility.True, out FOutputIndices);
            
		}

		#endregion pin creation
		

        #region mainloop

        public void Configurate(IPluginConfig Input)
        {
            //nothing to configure in this plugin
            //only used in conjunction with inputs of type cmpdConfigurate
        }

        //here we go, thats the method called by vvvv each frame
        //all data handling should be in here
        public void Evaluate(int SpreadMax)
        {
            //if any of the inputs has changed
            //recompute the outputs
            if (FInputVertices.PinIsChanged || FInputIndices.PinIsChanged || FInputNormals.PinIsChanged || FInputTexCoords.PinIsChanged)
            {
               // if (!FInputVertices.IsConnected || !FInputIndices.IsConnected || !FInputNormals.IsConnected || !FInputHeight.IsConnected || !FInputTexCoords.IsConnected)
                 //   return;

                LinkedList<Edge> edgeList = new LinkedList<Edge>();
                LinkedList<Vector3D> vertices = new LinkedList<Vector3D>();
                LinkedList<Vector3D> normals = new LinkedList<Vector3D>();
                LinkedList<Vector2D> texCoords = new LinkedList<Vector2D>();
                LinkedList<Vector2D> originalTexCoords = new LinkedList<Vector2D>();
                LinkedList<int> indices = new LinkedList<int>();
                //Dictionary<int, int> indexOldToNew = new Dictionary<int, int>();

                
                for (int face = 0; face < FInputIndices.SliceCount / 3; face++)
                {
                    double index1, index2, index3;
                    Vector3D v1 = new Vector3D(), v2 = new Vector3D(), v3 = new Vector3D();
                    Vector3D n1 = new Vector3D(), n2 = new Vector3D(), n3 = new Vector3D();
                    Vector2D t1 = new Vector2D(), t2 = new Vector2D(), t3 = new Vector2D();

                    FInputIndices.GetValue(face * 3, out index1);
                    FInputVertices.GetValue3D((int)index1, out v1.x, out v1.y, out v1.z);
                    FInputNormals.GetValue3D((int)index1, out n1.x, out n1.y, out n1.z);
                    FInputTexCoords.GetValue2D((int)index1, out t1.x, out t1.y);

                    FInputIndices.GetValue(face * 3 + 1, out index2);
                    FInputVertices.GetValue3D((int)index2, out v2.x, out v2.y, out v2.z);
                    FInputNormals.GetValue3D((int)index2, out n2.x, out n2.y, out n2.z);
                    FInputTexCoords.GetValue2D((int)index2, out t2.x, out t2.y);

                    FInputIndices.GetValue(face * 3 + 2, out index3);
                    FInputVertices.GetValue3D((int)index3, out v3.x, out v3.y, out v3.z);
                    FInputNormals.GetValue3D((int)index3, out n3.x, out n3.y, out n3.z);
                    FInputTexCoords.GetValue2D((int)index3, out t3.x, out t3.y);

                    n1 = ~n1;       // Normalize
                    n2 = ~n2;
                    n3 = ~n3;

                    Edge e = new Edge(v1, v2);
                    
                    if (edgeList.Find(e) == null)
                    {
                        edgeList.AddLast(e);
                        int iV1, iV2;
                        
                        iV1 = vertices.Count;
                        vertices.AddLast(v1);       // Add "to be extended vertex" directly after the original vertex
                        normals.AddLast(n1);
                        texCoords.AddLast(new Vector2D(0.0, 1.0));
                        originalTexCoords.AddLast(t1);
                        vertices.AddLast(v1);
                        normals.AddLast(n1);
                        texCoords.AddLast(new Vector2D(0.0, 0.0));
                        originalTexCoords.AddLast(t1);

                        iV2 = vertices.Count;
                        vertices.AddLast(v2);       // Add "to be extended vertex" directly after the original vertex
                        normals.AddLast(n2);
                        texCoords.AddLast(new Vector2D(1.0, 1.0));
                        originalTexCoords.AddLast(t2);
                        vertices.AddLast(v2);
                        normals.AddLast(n2);
                        texCoords.AddLast(new Vector2D(1.0, 0.0));
                        originalTexCoords.AddLast(t2);

                        indices.AddLast(iV1 + 1);
                        indices.AddLast(iV2);
                        indices.AddLast(iV1);

                        indices.AddLast(iV1 + 1);
                        indices.AddLast(iV2 + 1);
                        indices.AddLast(iV2);

                    }

                    e = new Edge(v2, v3);
                   

                    if (edgeList.Find(e) == null)
                    {
                        edgeList.AddLast(e);
                        int iV2, iV3;

                        iV2 = vertices.Count;
                        vertices.AddLast(v2);       // Add "to be extended vertex" directly after the original vertex
                        normals.AddLast(n2);
                        texCoords.AddLast(new Vector2D(0.0, 1.0));
                        originalTexCoords.AddLast(t2);
                        vertices.AddLast(v2);
                        normals.AddLast(n2);
                        texCoords.AddLast(new Vector2D(0.0, 0.0));
                        originalTexCoords.AddLast(t2);

                        iV3 = vertices.Count;
                        vertices.AddLast(v3);       // Add "to be extended vertex" directly after the original vertex
                        normals.AddLast(n3);
                        texCoords.AddLast(new Vector2D(1.0, 1.0));
                        originalTexCoords.AddLast(t3);
                        vertices.AddLast(v3);
                        normals.AddLast(n3);
                        texCoords.AddLast(new Vector2D(1.0, 0.0));
                        originalTexCoords.AddLast(t3);

                        indices.AddLast(iV2 + 1);
                        indices.AddLast(iV3);
                        indices.AddLast(iV2);

                        indices.AddLast(iV2 + 1);
                        indices.AddLast(iV3 + 1);
                        indices.AddLast(iV3);
                    }

                    e = new Edge(v3, v1);
                    

                    if (edgeList.Find(e) == null)
                    {
                        edgeList.AddLast(e);
                        int iV3, iV1;

                        iV3 = vertices.Count;
                        vertices.AddLast(v3);       // Add "to be extended vertex" directly after the original vertex
                        normals.AddLast(n3);
                        texCoords.AddLast(new Vector2D(0.0, 1.0));
                        originalTexCoords.AddLast(t3);
                        vertices.AddLast(v3);
                        normals.AddLast(n3);
                        texCoords.AddLast(new Vector2D(0.0, 0.0));
                        originalTexCoords.AddLast(t3);

                        iV1 = vertices.Count;
                        vertices.AddLast(v1);       // Add "to be extended vertex" directly after the original vertex
                        normals.AddLast(n1);
                        texCoords.AddLast(new Vector2D(1.0, 1.0));
                        originalTexCoords.AddLast(t1);
                        vertices.AddLast(v1);
                        normals.AddLast(n1);
                        texCoords.AddLast(new Vector2D(1.0, 0.0));
                        originalTexCoords.AddLast(t1);


                        indices.AddLast(iV3 + 1);
                        indices.AddLast(iV1);
                        indices.AddLast(iV3);

                        indices.AddLast(iV3 + 1);
                        indices.AddLast(iV1 + 1);
                        indices.AddLast(iV1);
                    }
                    
                }

                FOutputVertices.SliceCount = vertices.Count;        // ==
                FOutputNormals.SliceCount = normals.Count;
                FOutputTexCoords.SliceCount = texCoords.Count;
                FOutputOriginalTexCoords.SliceCount = originalTexCoords.Count;

                IEnumerator<Vector3D> vertEnumer = vertices.GetEnumerator();
                IEnumerator<Vector3D> normEnumer = normals.GetEnumerator();
                IEnumerator<Vector2D> texCEnumer = texCoords.GetEnumerator();
                IEnumerator<Vector2D> oTexCEnumer = originalTexCoords.GetEnumerator();
                for (int i = 0; vertEnumer.MoveNext() && texCEnumer.MoveNext() && oTexCEnumer.MoveNext() && normEnumer.MoveNext(); i++)
                {
                    FOutputVertices.SetValue3D(i, vertEnumer.Current.x, vertEnumer.Current.y, vertEnumer.Current.z);
                    FOutputNormals.SetValue3D(i, normEnumer.Current.x, normEnumer.Current.y, normEnumer.Current.z);
                    FOutputTexCoords.SetValue2D(i, texCEnumer.Current.x, texCEnumer.Current.y);
                    FOutputOriginalTexCoords.SetValue2D(i, oTexCEnumer.Current.x, oTexCEnumer.Current.y);
                }

                FOutputIndices.SliceCount = indices.Count;
                IEnumerator<int> indEnumer = indices.GetEnumerator();
                for (int index = 0; indEnumer.MoveNext() ; index++)
                    FOutputIndices.SetValue(index, indEnumer.Current);
            }
        }

        #endregion mainloop 
		
	}
}
